import json
import requests
from python_graphql_client import GraphqlClient

from verifit.login import get_bearer_auth_base


QUERY = 'query'
VARIABLES = 'variables'
METHOD = 'method'
LOG_PREFIX = 'logPrefix'
PAYLOAD = 'payload'


def retrieveGraphQl(url):
    def with_config(config):
        client = GraphqlClient(endpoint=url, auth=get_bearer_auth_base())
        prefix_message = config.get(LOG_PREFIX, 'Perform a POST request')
        prefix = f"{prefix_message} to {url}"
        print(prefix, 'variables', config[VARIABLES])
        response = client.execute(config[QUERY], config[VARIABLES])
        print(prefix, 'response', response)
        return response
    return with_config


def retrieveHttp(url):
    def with_config(config):
        method = config.get(METHOD, requests.get)
        prefix_message = config.get(LOG_PREFIX, f"Perform a {method} request")
        prefix = f"{prefix_message} to {url}"
        payload = config.get(PAYLOAD, None)
        print(prefix, 'payload', payload)
        response = method(
            url=url,
            data=None if payload is None else json.dumps(payload),
            headers={
                'Content-Type': 'application/json'
            },
        )
        print(prefix, 'response', response)
        data = response.json()
        print(prefix, 'JSON response', data)
        return data
    return with_config



from datetime import datetime

from requests.auth import AuthBase

from verifit.prop import get_prop

from .driver import get_driver

from .cache import cache_set, cache_get
from .config import get_store_reader, get_store_writer
from .date_and_time import date_diff_in_minutes
from .iam_token import decode_token, get_token_expiration_date

get_env = get_store_reader()
set_env = get_store_writer()

ACCESS_TOKEN = 'accessToken'
EXPIRY_DATE = 'expiryDate'
LOGIN_DATA = 'loginData'
USERNAME = 'username'
PASSWORD = 'password'


def get_expiry_date(login_data):
    return datetime.fromisoformat(get_prop(login_data)(EXPIRY_DATE))


def login(user):
    def with_driver(driver):
        access_token = driver(user)
        decoded_token = decode_token(access_token)
        token_expiry_date = get_token_expiration_date(decoded_token)
        login_data = {
            ACCESS_TOKEN: access_token,
            EXPIRY_DATE: token_expiry_date.isoformat()
        }
        cache_set(get_prop(user)(USERNAME))(login_data)
        set_env(LOGIN_DATA)(login_data)
        return login_data
    return with_driver


def login_main_user(driver_name):
    user = get_driver(driver_name)('user')('MAIN')
    return login(user)(get_driver(driver_name)('login'))


def login_main_user_from_cache(driver_name):
    user = get_driver(driver_name)('user')('MAIN')
    return login_from_cache(user)(get_driver(driver_name)('login'))


def get_login_values_from_cache(username):
    user_cached_data = cache_get(username)
    if user_cached_data is None:
        print(f"Cache miss key <{username}>, no cached data")
        return None
    token_expiry_iso_string = user_cached_data.get('expiryDate', None)
    if token_expiry_iso_string is None:
        print(f"Cache miss key <{username}>, no expiration info")
        return None
    token_expiry_date = datetime.fromisoformat(token_expiry_iso_string)
    minutes = date_diff_in_minutes(token_expiry_date)(datetime.now())
    if minutes < 10:
        print(f"Cache miss key <{username}>, almost expired")
        return None
    print(f"Cache hit key <{username}>")
    login_data = {
        ACCESS_TOKEN: get_prop(user_cached_data)(ACCESS_TOKEN),
        EXPIRY_DATE: token_expiry_date
    }
    return login_data


def login_from_cache(user):
    def with_driver(driver):
        login_data = get_login_values_from_cache(get_prop(user)(USERNAME))
        if login_data is None:
            login_data = login(user)(driver)
        print(f"Caching login data <{login_data}>")
        set_env(LOGIN_DATA)(login_data)
        return login_data
    return with_driver


def get_bearer_authorization_header_value():
    login_data = get_env(LOGIN_DATA)
    print(f"Using cached login data <{login_data}>")
    return f"Bearer {get_prop(login_data)(ACCESS_TOKEN)}"


def get_bearer_auth_base():
    class BearerAuth(AuthBase):
        def __init__(self, token):
            self.token = token

        def __call__(self, r):
            r.headers["authorization"] = "Bearer " + self.token
            return r

    login_data = get_env(LOGIN_DATA)
    return BearerAuth(get_prop(login_data)(ACCESS_TOKEN))



import asyncio
import json

import websockets

from .config import get_store_reader

get_env = get_store_reader()

SERVER = 'server'
IGNORE_LIST = 'ignoreList'


def ws_send_and_receive(config):
    server = config.get(SERVER)
    ignore_list = config.get(IGNORE_LIST)

    log_type = "[WebSockets]"
    wait_ms = int(get_env('WEBSOCKETS_WAIT_MS'))

    def with_data(data):
        async def receive_async():
            async with websockets.connect(server) as websocket:
                received = False
                while not received:
                    received_data = await websocket.recv()
                    if received_data in ignore_list:
                        print(f"Ignoring received message <{received_data}>")
                    else:
                        received = True
            return received_data

        async def receive_async_with_timeout():
            seconds = wait_ms / 1000
            received_data = await asyncio.wait_for(
                receive_async(),
                seconds
            )
            print(f"{log_type} Received <{received_data}>")
            return received_data

        def start_receiving():
            print(f"{log_type} Waiting {wait_ms} ms for packets from {server}")
            return asyncio.ensure_future(receive_async_with_timeout())

        def end_receiving(a_handle):
            received_data = asyncio.get_event_loop().run_until_complete(a_handle)
            print(f"{log_type} Finished waiting for packets from {server}, received {received_data}")
            return json.loads(received_data)

        async def send_async(data_to_send):
            async with websockets.connect(server) as websocket:
                await websocket.send(data_to_send)
                print(f"Sent <{data_to_send}>")

        def send_sync(data_to_send):
            asyncio.get_event_loop().run_until_complete(send_async(data_to_send))

        handle = start_receiving()
        to_send = json.dumps(data)
        send_sync(to_send)
        return end_receiving(handle)
    return with_data



def get_defaulted_prop(a_dict):
    def with_key(key):
        def with_default(default):
            return a_dict.get(key, default)
        return with_default
    return with_key


def get_prop(a_dict):
    def with_key(key):
        return get_defaulted_prop(a_dict)(key)(default=None)
    return with_key
